#include <avr/pgmspace.h>

#include "music_data.h"

#ifndef TRUE
#define TRUE  1
#endif
#ifndef FALSE
#define FALSE 0
#endif

//makes values in volume table into range of 0..84, for proper 8-bit stereo mixer limiting (84+84/2=126)

#define VOLUME_TABLE_DIVIDER  (780/2)

#include "AYEmu.h"
#include "PT3Play.h"

#define SAMPLE_RATE 31250 //half the PWM rate

volatile uint8_t frame_sync = 0;
volatile uint16_t frame_sample_cnt = 0;
volatile uint8_t out_pwm_l = 0;
volatile uint8_t out_pwm_r = 0;

PT3_SongInfo song;


void profile() {
  Serial.begin(115200);
  Serial.println("profiling start");
  uint32_t iterations = 1000000;
  uint16_t frame_samples = 0;
  uint32_t start_ms = millis();
  for (uint32_t i = 0; i < iterations; ++i)
  {
    AY_EmuAdvanceSample();
    ++frame_samples;
    if (frame_samples >= (SAMPLE_RATE / 50))
    {
      frame_samples = 0;
      PT3_PlayerUpdate(song);
      PT3_PlayerWriteAYRegs(song);
    }
  }
  uint32_t elapsed_ms = millis() - start_ms;
  char buf[128];
  sprintf(buf, "%lu iterations in %lu ms", iterations, elapsed_ms);
  Serial.println(buf);
  sprintf(buf, "max sample rate ~%lu Hz", iterations * 1000 / elapsed_ms);
  Serial.println(buf);
  Serial.end();
}



void setup() {

  AY_EmuInit(3500000 / 2, SAMPLE_RATE);

  PT3_Init(song, snowball_game);

 // profile();

  cli();

  // Set up Timer 1 for fast PWM mode (8-bit) on pin D9 (left channel)
  TCCR1A = (1 << COM1A1) | (1 << WGM10); // Fast PWM 8-bit, clear OC1A on compare match
  TCCR1B = (1 << WGM12) | (1 << CS10);   // No prescaling (full clock speed, 16 MHz)
  pinMode(9, OUTPUT);

  // Set up Timer 2 for fast PWM mode (8-bit) on pin D3 (right channel)
  TCCR2A = (1 << COM2B1) | (1 << WGM21) | (1 << WGM20); // Fast PWM, clear OC2B on compare match
  TCCR2B = (1 << CS20); // No prescaling (full clock speed, 16 MHz)
  pinMode(3, OUTPUT);

  // Set up Timer 0 for 31.25 kHz sample rate interrupt
  TCCR0A = (1 << WGM01);  //CTC mode
  TCCR0B = (1 << CS01); // Prescaler = 8
  OCR0A = (16000000 / 8 / (SAMPLE_RATE)) - 1; // Compare value for 31.25 kHz
  TIMSK0 = (1 << OCIE0A); // Enable compare match interrupt

  sei();
}


// Timer 0 compare match interrupt (sample rate)
ISR(TIMER0_COMPA_vect) {

  // Update the PWM duty cycles
  OCR1A = out_pwm_l;;  // Left channel (pin D9)
  OCR2B = out_pwm_r; // Right channel (pin D3)

  AY_EmuAdvanceSample();

  uint8_t a = AY_EmuVolumeTable[AY_EmuGetDAC_A()];
  uint8_t b = AY_EmuVolumeTable[AY_EmuGetDAC_B()];
  uint8_t c = AY_EmuVolumeTable[AY_EmuGetDAC_C()];

  out_pwm_l = a + b / 2;
  out_pwm_r = c + b / 2;

  ++frame_sample_cnt;

  if (frame_sample_cnt >= (SAMPLE_RATE / 50))
  {
    frame_sample_cnt = 0;
    ++frame_sync;
  }
}



void loop() {
  while (1)
  {
    uint8_t sync_prev = frame_sync;

    while (frame_sync == sync_prev);

    PT3_PlayerUpdate(song);
    cli();
    PT3_PlayerWriteAYRegs(song);
    sei();
  }
}
